/*
**  PROGRAM: Parallel Matrix Multiply (using OpenMP)
**
**  PURPOSE: This is a simple matrix multiply program. 
**           It will compute the product
**
**                C  = A * B
**
**           A and B are set to constant matrices so we
**           can make a quick test of the multiplication.
**
**  USAGE:   Right now, I hardwire the martix dimensions. 
**           later, I'll take them from the command line.
**  
**  HISTORY: Written by Tim Mattson, Nov 1999. Modified by T. Grenier 11/2015
*/
#include <iostream>
#include <omp.h>
using namespace std;


#define Nthreads 1

typedef double DataType;
#define ORDER 2000  // Matrices size. 2000 and DataType as double will need arround 60s with one thread


#define AVAL 3.0
#define BVAL 5.0
#define TOL  0.001

int main(int argc, char *argv[])
{
	int Ndim, Pdim, Mdim;   /* A[N][P], B[P][M], C[N][M] */
	int i,j,k;
	
	DataType *A, *B, *C, cval, tmp;
	double err, errsq;
    double dN, mflops;
	double start_time, run_time;


	Ndim = ORDER;
	Pdim = ORDER;
	Mdim = ORDER;

	A = new DataType[Ndim*Pdim];
    B = new DataType[Pdim*Mdim];
    C = new DataType[Ndim*Mdim];

	/* Initialize matrices */

	for (i=0; i<Ndim; i++)
		for (j=0; j<Pdim; j++)
			*(A+(i*Ndim+j)) = AVAL;

	for (i=0; i<Pdim; i++)
		for (j=0; j<Mdim; j++)
			*(B+(i*Pdim+j)) = BVAL;

	for (i=0; i<Ndim; i++)
		for (j=0; j<Mdim; j++)
			*(C+(i*Ndim+j)) = 0.0;
	
	omp_set_num_threads(Nthreads);
	
		start_time = omp_get_wtime();

	/* Do the matrix product */

#pragma omp parallel for private(tmp, i, j, k)  
	for (i=0; i<Ndim; i++){
		for (j=0; j<Mdim; j++){
 
            tmp = 0.0;

			for(k=0;k<Pdim;k++){
				/* C(i,j) = sum(over k) A(i,k) * B(k,j) */
				//tmp += *(A+(i*Ndim+k)) *  *(B+(k*Pdim+j));
				 tmp+= A[i*Ndim+k] * B[k*Pdim+j];
			}
			//*(C+(i*Ndim+j)) = tmp;
			C[i*Ndim+j] = tmp;
		}
	}
	
	
	/* Check the answer */
	run_time = omp_get_wtime() - start_time;

	cout << " Order " << ORDER << " multiplications in " << run_time << " seconds \n";
    cout << omp_get_max_threads() <<" threads\n";
    dN = (double)ORDER;
    mflops = 2.0 * dN * dN * dN/(1000000.0* run_time);

     cout << " Order " << ORDER << " multiplications at " << mflops << " mflops\n";

	cval = Pdim * AVAL * BVAL;
	errsq = 0.0;
	for (i=0; i<Ndim; i++){
		for (j=0; j<Mdim; j++){
			err = *(C+i*Ndim+j) - cval;
		    errsq += err * err;
		}
	}

	if (errsq > TOL) 
		cout << "\n Errors in multiplication: " << errsq << endl;
	else
		cout << "\n Hey, it worked" << endl;;

	
	delete[] A;
	delete[] B;
	delete[] C;
	
	return 0;
}
